@@ -31,6 +31,7 @@ from cocotb.triggers import Event
 from .version import __version__
 from .constants import AxiProt, AxiResp
 from .axil_channels import AxiLiteAWSource, AxiLiteWSource, AxiLiteBSink, AxiLiteARSource, AxiLiteRSink
+from .reset import Reset
 
 # AXI lite master write
 AxiLiteWriteCmd = namedtuple("AxiLiteWriteCmd", ["address", "data", "prot", "event"])
@@ -43,31 +44,30 @@ AxiLiteReadRespCmd = namedtuple("AxiLiteReadRespCmd", ["address", "length", "cyc
 AxiLiteReadResp = namedtuple("AxiLiteReadResp", ["address", "data", "resp"])
 
 
-class AxiLiteMasterWrite(object):
-    def __init__(self, entity, name, clock, reset=None):
-        self.log = logging.getLogger(f"cocotb.{entity._name}.{name}")
+class AxiLiteMasterWrite(Reset):
+    def __init__(self, bus, clock, reset=None, reset_active_level=True):
+        self.log = logging.getLogger(f"cocotb.{bus.aw._entity._name}.{bus.aw._name}")
 
         self.log.info("AXI lite master (write)")
         self.log.info("cocotbext-axi version %s", __version__)
         self.log.info("Copyright (c) 2020 Alex Forencich")
         self.log.info("https://github.com/alexforencich/cocotbext-axi")
 
-        self.reset = reset
-
-        self.aw_channel = AxiLiteAWSource(entity, name, clock, reset)
-        self.w_channel = AxiLiteWSource(entity, name, clock, reset)
-        self.b_channel = AxiLiteBSink(entity, name, clock, reset)
+        self.aw_channel = AxiLiteAWSource(bus.aw, clock, reset, reset_active_level)
+        self.w_channel = AxiLiteWSource(bus.w, clock, reset, reset_active_level)
+        self.b_channel = AxiLiteBSink(bus.b, clock, reset, reset_active_level)
 
         self.write_command_queue = deque()
         self.write_command_sync = Event()
         self.write_resp_queue = deque()
         self.write_resp_sync = Event()
-        self.write_resp_set = set()
 
         self.int_write_resp_command_queue = deque()
         self.int_write_resp_command_sync = Event()
 
         self.in_flight_operations = 0
+        self._idle = Event()
+        self._idle.set()
 
         self.width = len(self.w_channel.bus.wdata)
         self.byte_size = 8
@@ -82,14 +82,17 @@ class AxiLiteMasterWrite(object):
         assert self.byte_width == len(self.w_channel.bus.wstrb)
         assert self.byte_width * self.byte_size == self.width
 
-        cocotb.fork(self._process_write())
-        cocotb.fork(self._process_write_resp())
+        self._process_write_cr = None
+        self._process_write_resp_cr = None
+
+        self._init_reset(reset, reset_active_level)
 
     def init_write(self, address, data, prot=AxiProt.NONSECURE, event=None):
         if event is not None and not isinstance(event, Event):
             raise ValueError("Expected event object")
 
         self.in_flight_operations += 1
+        self._idle.clear()
 
         self.write_command_queue.append(AxiLiteWriteCmd(address, bytearray(data), prot, event))
         self.write_command_sync.set()
@@ -99,8 +102,7 @@ class AxiLiteMasterWrite(object):
 
     async def wait(self):
         while not self.idle():
-            self.write_resp_sync.clear()
-            await self.write_resp_sync.wait()
+            await self._idle.wait()
 
     def write_resp_ready(self):
         return bool(self.write_resp_queue)
@@ -141,6 +143,41 @@ class AxiLiteMasterWrite(object):
     async def write_qword(self, address, data, byteorder='little', prot=AxiProt.NONSECURE):
         await self.write_qwords(address, [data], byteorder, prot)
 
+    def _handle_reset(self, state):
+        if state:
+            self.log.info("Reset asserted")
+            if self._process_write_cr is not None:
+                self._process_write_cr.kill()
+                self._process_write_cr = None
+            if self._process_write_resp_cr is not None:
+                self._process_write_resp_cr.kill()
+                self._process_write_resp_cr = None
+        else:
+            self.log.info("Reset de-asserted")
+            if self._process_write_cr is None:
+                self._process_write_cr = cocotb.fork(self._process_write())
+            if self._process_write_resp_cr is None:
+                self._process_write_resp_cr = cocotb.fork(self._process_write_resp())
+
+        self.aw_channel.clear()
+        self.w_channel.clear()
+        self.b_channel.clear()
+
+        while self.write_command_queue:
+            cmd = self.write_command_queue.popleft()
+            if cmd.event:
+                cmd.event.set(None)
+
+        while self.int_write_resp_command_queue:
+            cmd = self.int_write_resp_command_queue.popleft()
+            if cmd.event:
+                cmd.event.set(None)
+
+        self.write_resp_queue.clear()
+
+        self.in_flight_operations = 0
+        self._idle.set()
+
     async def _process_write(self):
         while True:
             if not self.write_command_queue:
@@ -193,7 +230,7 @@ class AxiLiteMasterWrite(object):
                 w.wdata = val
                 w.wstrb = strb
 
-                await self.aw_channel.drive(aw)
+                await self.aw_channel.send(aw)
                 await self.w_channel.send(w)
 
     async def _process_write_resp(self):
@@ -227,31 +264,33 @@ class AxiLiteMasterWrite(object):
 
             self.in_flight_operations -= 1
 
+            if self.in_flight_operations == 0:
+                self._idle.set()
 
-class AxiLiteMasterRead(object):
-    def __init__(self, entity, name, clock, reset=None):
-        self.log = logging.getLogger(f"cocotb.{entity._name}.{name}")
+
+class AxiLiteMasterRead(Reset):
+    def __init__(self, bus, clock, reset=None, reset_active_level=True):
+        self.log = logging.getLogger(f"cocotb.{bus.ar._entity._name}.{bus.ar._name}")
 
         self.log.info("AXI lite master (read)")
         self.log.info("cocotbext-axi version %s", __version__)
         self.log.info("Copyright (c) 2020 Alex Forencich")
         self.log.info("https://github.com/alexforencich/cocotbext-axi")
 
-        self.reset = reset
-
-        self.ar_channel = AxiLiteARSource(entity, name, clock, reset)
-        self.r_channel = AxiLiteRSink(entity, name, clock, reset)
+        self.ar_channel = AxiLiteARSource(bus.ar, clock, reset, reset_active_level)
+        self.r_channel = AxiLiteRSink(bus.r, clock, reset, reset_active_level)
 
         self.read_command_queue = deque()
         self.read_command_sync = Event()
         self.read_data_queue = deque()
         self.read_data_sync = Event()
-        self.read_data_set = set()
 
         self.int_read_resp_command_queue = deque()
         self.int_read_resp_command_sync = Event()
 
         self.in_flight_operations = 0
+        self._idle = Event()
+        self._idle.set()
 
         self.width = len(self.r_channel.bus.rdata)
         self.byte_size = 8
@@ -264,14 +303,17 @@ class AxiLiteMasterRead(object):
 
         assert self.byte_width * self.byte_size == self.width
 
-        cocotb.fork(self._process_read())
-        cocotb.fork(self._process_read_resp())
+        self._process_read_cr = None
+        self._process_read_resp_cr = None
+
+        self._init_reset(reset, reset_active_level)
 
     def init_read(self, address, length, prot=AxiProt.NONSECURE, event=None):
         if event is not None and not isinstance(event, Event):
             raise ValueError("Expected event object")
 
         self.in_flight_operations += 1
+        self._idle.clear()
 
         self.read_command_queue.append(AxiLiteReadCmd(address, length, prot, event))
         self.read_command_sync.set()
@@ -281,8 +323,7 @@ class AxiLiteMasterRead(object):
 
     async def wait(self):
         while not self.idle():
-            self.read_data_sync.clear()
-            await self.read_data_sync.wait()
+            await self._idle.wait()
 
     def read_data_ready(self):
         return bool(self.read_data_queue)
@@ -323,6 +364,40 @@ class AxiLiteMasterRead(object):
     async def read_qword(self, address, byteorder='little', prot=AxiProt.NONSECURE):
         return (await self.read_qwords(address, 1, byteorder, prot))[0]
 
+    def _handle_reset(self, state):
+        if state:
+            self.log.info("Reset asserted")
+            if self._process_read_cr is not None:
+                self._process_read_cr.kill()
+                self._process_read_cr = None
+            if self._process_read_resp_cr is not None:
+                self._process_read_resp_cr.kill()
+                self._process_read_resp_cr = None
+        else:
+            self.log.info("Reset de-asserted")
+            if self._process_read_cr is None:
+                self._process_read_cr = cocotb.fork(self._process_read())
+            if self._process_read_resp_cr is None:
+                self._process_read_resp_cr = cocotb.fork(self._process_read_resp())
+
+        self.ar_channel.clear()
+        self.r_channel.clear()
+
+        while self.read_command_queue:
+            cmd = self.read_command_queue.popleft()
+            if cmd.event:
+                cmd.event.set(None)
+
+        while self.int_read_resp_command_queue:
+            cmd = self.int_read_resp_command_queue.popleft()
+            if cmd.event:
+                cmd.event.set(None)
+
+        self.read_data_queue.clear()
+
+        self.in_flight_operations = 0
+        self._idle.set()
+
     async def _process_read(self):
         while True:
             if not self.read_command_queue:
@@ -347,7 +422,7 @@ class AxiLiteMasterRead(object):
                 ar.araddr = word_addr + k*self.byte_width
                 ar.arprot = cmd.prot
 
-                await self.ar_channel.drive(ar)
+                await self.ar_channel.send(ar)
 
     async def _process_read_resp(self):
         while True:
@@ -397,14 +472,17 @@ class AxiLiteMasterRead(object):
 
             self.in_flight_operations -= 1
 
+            if self.in_flight_operations == 0:
+                self._idle.set()
 
-class AxiLiteMaster(object):
-    def __init__(self, entity, name, clock, reset=None):
+
+class AxiLiteMaster:
+    def __init__(self, bus, clock, reset=None, reset_active_level=True):
         self.write_if = None
         self.read_if = None
 
-        self.write_if = AxiLiteMasterWrite(entity, name, clock, reset)
-        self.read_if = AxiLiteMasterRead(entity, name, clock, reset)
+        self.write_if = AxiLiteMasterWrite(bus.write, clock, reset, reset_active_level)
+        self.read_if = AxiLiteMasterRead(bus.read, clock, reset, reset_active_level)
 
     def init_read(self, address, length, prot=AxiProt.NONSECURE, event=None):
         self.read_if.init_read(address, length, prot, event)
