@@ -26,19 +26,24 @@ import logging
 from collections import deque
 
 import cocotb
-from cocotb.triggers import RisingEdge, ReadOnly, Timer, First, Event
+from cocotb.triggers import RisingEdge, Timer, First, Event
+from cocotb.utils import get_sim_time
 from cocotb.bus import Bus
 
 from .version import __version__
+from .reset import Reset
 
 
-class AxiStreamFrame(object):
-    def __init__(self, tdata=b'', tkeep=None, tid=None, tdest=None, tuser=None):
+class AxiStreamFrame:
+    def __init__(self, tdata=b'', tkeep=None, tid=None, tdest=None, tuser=None, tx_complete=None):
         self.tdata = bytearray()
         self.tkeep = None
         self.tid = None
         self.tdest = None
         self.tuser = None
+        self.sim_time_start = None
+        self.sim_time_end = None
+        self.tx_complete = None
 
         if type(tdata) is AxiStreamFrame:
             if type(tdata.tdata) is bytearray:
@@ -62,6 +67,9 @@ class AxiStreamFrame(object):
                     self.tuser = tdata.tuser
                 else:
                     self.tuser = list(tdata.tuser)
+            self.sim_time_start = tdata.sim_time_start
+            self.sim_time_end = tdata.sim_time_end
+            self.tx_complete = tdata.tx_complete
         elif type(tdata) in (bytes, bytearray):
             self.tdata = bytearray(tdata)
             self.tkeep = tkeep
@@ -75,6 +83,9 @@ class AxiStreamFrame(object):
             self.tdest = tdest
             self.tuser = tuser
 
+        if tx_complete is not None:
+            self.tx_complete = tx_complete
+
     def normalize(self):
         # normalize all sideband signals to the same size as tdata
         n = len(self.tdata)
@@ -144,6 +155,12 @@ class AxiStreamFrame(object):
         elif all(self.tuser[0] == i for i in self.tuser):
             self.tuser = self.tuser[0]
 
+    def handle_tx_complete(self):
+        if isinstance(self.tx_complete, Event):
+            self.tx_complete.set(self)
+        elif callable(self.tx_complete):
+            self.tx_complete(self)
+
     def __eq__(self, other):
         if not isinstance(other, AxiStreamFrame):
             return False
@@ -195,11 +212,13 @@ class AxiStreamFrame(object):
 
     def __repr__(self):
         return (
-            f"{type(self).__name__}(tdata={repr(self.tdata)}, "
-            f"tkeep={repr(self.tkeep)}, "
-            f"tid={repr(self.tid)}, "
-            f"tdest={repr(self.tdest)}, "
-            f"tuser={repr(self.tuser)})"
+            f"{type(self).__name__}(tdata={self.tdata!r}, "
+            f"tkeep={self.tkeep!r}, "
+            f"tid={self.tid!r}, "
+            f"tdest={self.tdest!r}, "
+            f"tuser={self.tuser!r}, "
+            f"sim_time_start={self.sim_time_start!r}, "
+            f"sim_time_end={self.sim_time_end!r})"
         )
 
     def __len__(self):
@@ -212,19 +231,44 @@ class AxiStreamFrame(object):
         return bytes(self.tdata)
 
 
-class AxiStreamSource(object):
+class AxiStreamBus(Bus):
 
     _signals = ["tdata"]
     _optional_signals = ["tvalid", "tready", "tlast", "tkeep", "tid", "tdest", "tuser"]
 
-    def __init__(self, entity, name, clock, reset=None, byte_size=None, byte_lanes=None, *args, **kwargs):
-        self.log = logging.getLogger(f"cocotb.{entity._name}.{name}")
-        self.entity = entity
+    def __init__(self, entity=None, prefix=None, **kwargs):
+        super().__init__(entity, prefix, self._signals, optional_signals=self._optional_signals, **kwargs)
+
+    @classmethod
+    def from_entity(cls, entity, **kwargs):
+        return cls(entity, **kwargs)
+
+    @classmethod
+    def from_prefix(cls, entity, prefix, **kwargs):
+        return cls(entity, prefix, **kwargs)
+
+
+class AxiStreamBase(Reset):
+
+    _signals = ["tdata"]
+    _optional_signals = ["tvalid", "tready", "tlast", "tkeep", "tid", "tdest", "tuser"]
+
+    _type = "base"
+
+    _init_x = False
+
+    _valid_init = None
+    _ready_init = None
+
+    def __init__(self, bus, clock, reset=None, reset_active_level=True,
+            byte_size=None, byte_lanes=None, *args, **kwargs):
+
+        self.bus = bus
         self.clock = clock
         self.reset = reset
-        self.bus = Bus(self.entity, name, self._signals, optional_signals=self._optional_signals, **kwargs)
+        self.log = logging.getLogger(f"cocotb.{bus._entity._name}.{bus._name}")
 
-        self.log.info("AXI stream source")
+        self.log.info("AXI stream %s", self._type)
         self.log.info("cocotbext-axi version %s", __version__)
         self.log.info("Copyright (c) 2020 Alex Forencich")
         self.log.info("https://github.com/alexforencich/cocotbext-axi")
@@ -233,10 +277,7 @@ class AxiStreamSource(object):
 
         self.active = False
         self.queue = deque()
-
-        self.pause = False
-        self._pause_generator = None
-        self._pause_cr = None
+        self.queue_sync = Event()
 
         self.queue_occupancy_bytes = 0
         self.queue_occupancy_frames = 0
@@ -244,21 +285,17 @@ class AxiStreamSource(object):
         self.width = len(self.bus.tdata)
         self.byte_lanes = 1
 
-        self.reset = reset
+        if self._valid_init is not None and hasattr(self.bus, "tvalid"):
+            self.bus.tvalid.setimmediatevalue(self._valid_init)
+        if self._ready_init is not None and hasattr(self.bus, "tready"):
+            self.bus.tready.setimmediatevalue(self._ready_init)
 
-        self.bus.tdata.setimmediatevalue(0)
-        if hasattr(self.bus, "tvalid"):
-            self.bus.tvalid.setimmediatevalue(0)
-        if hasattr(self.bus, "tlast"):
-            self.bus.tlast.setimmediatevalue(0)
-        if hasattr(self.bus, "tkeep"):
-            self.bus.tkeep.setimmediatevalue(0)
-        if hasattr(self.bus, "tid"):
-            self.bus.tid.setimmediatevalue(0)
-        if hasattr(self.bus, "tdest"):
-            self.bus.tdest.setimmediatevalue(0)
-        if hasattr(self.bus, "tuser"):
-            self.bus.tuser.setimmediatevalue(0)
+        for sig in self._signals+self._optional_signals:
+            if hasattr(self.bus, sig):
+                if self._init_x and sig not in ("tvalid", "tready"):
+                    v = getattr(self.bus, sig).value
+                    v.binstr = 'x'*len(v)
+                    getattr(self.bus, sig).setimmediatevalue(v)
 
         if hasattr(self.bus, "tkeep"):
             self.byte_lanes = len(self.bus.tkeep)
@@ -275,7 +312,7 @@ class AxiStreamSource(object):
         self.byte_size = self.width // self.byte_lanes
         self.byte_mask = 2**self.byte_size-1
 
-        self.log.info("AXI stream source configuration:")
+        self.log.info("AXI stream %s configuration:", self._type)
         self.log.info("  Byte size: %d bits", self.byte_size)
         self.log.info("  Data width: %d bits (%d bytes)", self.width, self.byte_lanes)
         self.log.info("  tvalid: %s", "present" if hasattr(self.bus, "tvalid") else "not present")
@@ -302,7 +339,73 @@ class AxiStreamSource(object):
             raise ValueError(f"Bus does not evenly divide into byte lanes "
                 f"({self.byte_lanes} * {self.byte_size} != {self.width})")
 
-        cocotb.fork(self._run())
+        self._run_cr = None
+
+        self._init_reset(reset, reset_active_level)
+
+    def count(self):
+        return len(self.queue)
+
+    def empty(self):
+        return not self.queue
+
+    def clear(self):
+        self.queue.clear()
+        self.queue_occupancy_bytes = 0
+        self.queue_occupancy_frames = 0
+
+    def _handle_reset(self, state):
+        if state:
+            self.log.info("Reset asserted")
+            if self._run_cr is not None:
+                self._run_cr.kill()
+                self._run_cr = None
+        else:
+            self.log.info("Reset de-asserted")
+            if self._run_cr is None:
+                self._run_cr = cocotb.fork(self._run())
+
+        self.active = False
+
+    async def _run(self):
+        raise NotImplementedError()
+
+
+class AxiStreamPause:
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        self.pause = False
+        self._pause_generator = None
+        self._pause_cr = None
+
+    def set_pause_generator(self, generator=None):
+        if self._pause_cr is not None:
+            self._pause_cr.kill()
+            self._pause_cr = None
+
+        self._pause_generator = generator
+
+        if self._pause_generator is not None:
+            self._pause_cr = cocotb.fork(self._run_pause())
+
+    def clear_pause_generator(self):
+        self.set_pause_generator(None)
+
+    async def _run_pause(self):
+        for val in self._pause_generator:
+            self.pause = val
+            await RisingEdge(self.clock)
+
+
+class AxiStreamSource(AxiStreamBase, AxiStreamPause):
+
+    _type = "source"
+
+    _init_x = True
+
+    _valid_init = 0
+    _ready_init = None
 
     async def send(self, frame):
         self.send_nowait(frame)
@@ -319,12 +422,6 @@ class AxiStreamSource(object):
     def write_nowait(self, data):
         self.send_nowait(data)
 
-    def count(self):
-        return len(self.queue)
-
-    def empty(self):
-        return not self.queue
-
     def idle(self):
         return self.empty() and not self.active
 
@@ -332,56 +429,41 @@ class AxiStreamSource(object):
         while not self.idle():
             await RisingEdge(self.clock)
 
-    def set_pause_generator(self, generator=None):
-        if self._pause_cr is not None:
-            self._pause_cr.kill()
-            self._pause_cr = None
+    def _handle_reset(self, state):
+        super()._handle_reset(state)
 
-        self._pause_generator = generator
-
-        if self._pause_generator is not None:
-            self._pause_cr = cocotb.fork(self._run_pause())
-
-    def clear_pause_generator(self):
-        self.set_pause_generator(None)
+        self.bus.tdata <= 0
+        if hasattr(self.bus, "tvalid"):
+            self.bus.tvalid <= 0
+        if hasattr(self.bus, "tlast"):
+            self.bus.tlast <= 0
+        if hasattr(self.bus, "tkeep"):
+            self.bus.tkeep <= 0
+        if hasattr(self.bus, "tid"):
+            self.bus.tid <= 0
+        if hasattr(self.bus, "tdest"):
+            self.bus.tdest <= 0
+        if hasattr(self.bus, "tuser"):
+            self.bus.tuser <= 0
 
     async def _run(self):
         frame = None
         self.active = False
 
         while True:
-            await ReadOnly()
+            await RisingEdge(self.clock)
 
             # read handshake signals
             tready_sample = (not hasattr(self.bus, "tready")) or self.bus.tready.value
             tvalid_sample = (not hasattr(self.bus, "tvalid")) or self.bus.tvalid.value
 
-            if self.reset is not None and self.reset.value:
-                await RisingEdge(self.clock)
-                frame = None
-                self.active = False
-                self.bus.tdata <= 0
-                if hasattr(self.bus, "tvalid"):
-                    self.bus.tvalid <= 0
-                if hasattr(self.bus, "tlast"):
-                    self.bus.tlast <= 0
-                if hasattr(self.bus, "tkeep"):
-                    self.bus.tkeep <= 0
-                if hasattr(self.bus, "tid"):
-                    self.bus.tid <= 0
-                if hasattr(self.bus, "tdest"):
-                    self.bus.tdest <= 0
-                if hasattr(self.bus, "tuser"):
-                    self.bus.tuser <= 0
-                continue
-
-            await RisingEdge(self.clock)
-
             if (tready_sample and tvalid_sample) or not tvalid_sample:
                 if frame is None and self.queue:
                     frame = self.queue.popleft()
                     self.queue_occupancy_bytes -= len(frame)
                     self.queue_occupancy_frames -= 1
+                    frame.sim_time_start = get_sim_time()
+                    frame.sim_time_end = None
                     self.log.info("TX frame: %s", frame)
                     frame.normalize()
                     self.active = True
@@ -403,6 +485,8 @@ class AxiStreamSource(object):
 
                         if len(frame.tdata) == 0:
                             tlast_val = 1
+                            frame.sim_time_end = get_sim_time()
+                            frame.handle_tx_complete()
                             frame = None
                             break
 
@@ -426,101 +510,27 @@ class AxiStreamSource(object):
                         self.bus.tlast <= 0
                     self.active = bool(frame)
 
-    async def _run_pause(self):
-        for val in self._pause_generator:
-            self.pause = val
-            await RisingEdge(self.clock)
 
+class AxiStreamMonitor(AxiStreamBase):
 
-class AxiStreamSink(object):
+    _type = "monitor"
 
-    _signals = ["tdata"]
-    _optional_signals = ["tvalid", "tready", "tlast", "tkeep", "tid", "tdest", "tuser"]
+    _init_x = False
 
-    def __init__(self, entity, name, clock, reset=None, byte_size=None, byte_lanes=None, *args, **kwargs):
-        self.log = logging.getLogger(f"cocotb.{entity._name}.{name}")
-        self.entity = entity
-        self.clock = clock
-        self.reset = reset
-        self.bus = Bus(self.entity, name, self._signals, optional_signals=self._optional_signals, **kwargs)
+    _valid_init = None
+    _ready_init = None
 
-        self.log.info("AXI stream sink")
-        self.log.info("cocotbext-axi version %s", __version__)
-        self.log.info("Copyright (c) 2020 Alex Forencich")
-        self.log.info("https://github.com/alexforencich/cocotbext-axi")
+    def __init__(self, bus, clock, reset=None, reset_active_level=True,
+            byte_size=None, byte_lanes=None, *args, **kwargs):
 
-        super().__init__(*args, **kwargs)
+        super().__init__(bus, clock, reset, reset_active_level, byte_size, byte_lanes, *args, **kwargs)
 
-        self.active = False
-        self.queue = deque()
-        self.sync = Event()
         self.read_queue = []
 
-        self.pause = False
-        self._pause_generator = None
-        self._pause_cr = None
-
-        self.queue_occupancy_bytes = 0
-        self.queue_occupancy_frames = 0
-        self.queue_occupancy_limit_bytes = None
-        self.queue_occupancy_limit_frames = None
-
-        self.width = len(self.bus.tdata)
-        self.byte_lanes = 1
-
-        self.reset = reset
-
-        if hasattr(self.bus, "tready"):
-            self.bus.tready.setimmediatevalue(0)
-
-        if hasattr(self.bus, "tkeep"):
-            self.byte_lanes = len(self.bus.tkeep)
-            if byte_size is not None or byte_lanes is not None:
-                raise ValueError("Cannot specify byte_size or byte_lanes if tkeep is connected")
-        else:
-            if byte_lanes is not None:
-                self.byte_lanes = byte_lanes
-                if byte_size is not None:
-                    raise ValueError("Cannot specify both byte_size and byte_lanes")
-            elif byte_size is not None:
-                self.byte_lanes = self.width // byte_size
-
-        self.byte_size = self.width // self.byte_lanes
-        self.byte_mask = 2**self.byte_size-1
-
-        self.log.info("AXI stream sink configuration:")
-        self.log.info("  Byte size: %d bits", self.byte_size)
-        self.log.info("  Data width: %d bits (%d bytes)", self.width, self.byte_lanes)
-        self.log.info("  tvalid: %s", "present" if hasattr(self.bus, "tvalid") else "not present")
-        self.log.info("  tready: %s", "present" if hasattr(self.bus, "tready") else "not present")
-        self.log.info("  tlast: %s", "present" if hasattr(self.bus, "tlast") else "not present")
-        if hasattr(self.bus, "tkeep"):
-            self.log.info("  tkeep width: %d bits", len(self.bus.tkeep))
-        else:
-            self.log.info("  tkeep: not present")
-        if hasattr(self.bus, "tid"):
-            self.log.info("  tid width: %d bits", len(self.bus.tid))
-        else:
-            self.log.info("  tid: not present")
-        if hasattr(self.bus, "tdest"):
-            self.log.info("  tdest width: %d bits", len(self.bus.tdest))
-        else:
-            self.log.info("  tdest: not present")
-        if hasattr(self.bus, "tuser"):
-            self.log.info("  tuser width: %d bits", len(self.bus.tuser))
-        else:
-            self.log.info("  tuser: not present")
-
-        if self.byte_lanes * self.byte_size != self.width:
-            raise ValueError(f"Bus does not evenly divide into byte lanes "
-                f"({self.byte_lanes} * {self.byte_size} != {self.width})")
-
-        cocotb.fork(self._run())
-
     async def recv(self, compact=True):
         while self.empty():
-            self.sync.clear()
-            await self.sync.wait()
+            self.queue_sync.clear()
+            await self.queue_sync.wait()
         return self.recv_nowait(compact)
 
     def recv_nowait(self, compact=True):
@@ -549,65 +559,37 @@ class AxiStreamSink(object):
         del self.read_queue[:count]
         return data
 
-    def count(self):
-        return len(self.queue)
-
-    def empty(self):
-        return not self.queue
-
-    def full(self):
-        if self.queue_occupancy_limit_bytes and self.queue_occupancy_bytes > self.queue_occupancy_limit_bytes:
-            return True
-        elif self.queue_occupancy_limit_frames and self.queue_occupancy_frames > self.queue_occupancy_limit_frames:
-            return True
-        else:
-            return False
-
     def idle(self):
         return not self.active
 
     async def wait(self, timeout=0, timeout_unit='ns'):
         if not self.empty():
             return
-        self.sync.clear()
+        self.queue_sync.clear()
         if timeout:
-            await First(self.sync.wait(), Timer(timeout, timeout_unit))
+            await First(self.queue_sync.wait(), Timer(timeout, timeout_unit))
         else:
-            await self.sync.wait()
-
-    def set_pause_generator(self, generator=None):
-        if self._pause_cr is not None:
-            self._pause_cr.kill()
-            self._pause_cr = None
-
-        self._pause_generator = generator
-
-        if self._pause_generator is not None:
-            self._pause_cr = cocotb.fork(self._run_pause())
-
-    def clear_pause_generator(self):
-        self.set_pause_generator(None)
+            await self.queue_sync.wait()
 
     async def _run(self):
-        frame = AxiStreamFrame([], [], [], [], [])
+        frame = None
         self.active = False
 
         while True:
-            await ReadOnly()
+            await RisingEdge(self.clock)
 
             # read handshake signals
             tready_sample = (not hasattr(self.bus, "tready")) or self.bus.tready.value
             tvalid_sample = (not hasattr(self.bus, "tvalid")) or self.bus.tvalid.value
 
-            if self.reset is not None and self.reset.value:
-                await RisingEdge(self.clock)
-                frame = AxiStreamFrame([], [], [], [], [])
-                self.active = False
-                if hasattr(self.bus, "tready"):
-                    self.bus.tready <= 0
-                continue
-
             if tready_sample and tvalid_sample:
+                if frame is None:
+                    if self.byte_size == 8:
+                        frame = AxiStreamFrame(bytearray(), [], [], [], [])
+                    else:
+                        frame = AxiStreamFrame([], [], [], [], [])
+                    frame.sim_time_start = get_sim_time()
+
                 for offset in range(self.byte_lanes):
 
                     frame.tdata.append((self.bus.tdata.value.integer >> (offset * self.byte_size)) & self.byte_mask)
@@ -621,174 +603,68 @@ class AxiStreamSink(object):
                         frame.tuser.append(self.bus.tuser.value.integer)
 
                 if not hasattr(self.bus, "tlast") or self.bus.tlast.value:
-                    if self.byte_size == 8:
-                        frame.tdata = bytearray(frame.tdata)
-
+                    frame.sim_time_end = get_sim_time()
                     self.log.info("RX frame: %s", frame)
 
                     self.queue_occupancy_bytes += len(frame)
                     self.queue_occupancy_frames += 1
 
                     self.queue.append(frame)
-                    self.sync.set()
+                    self.queue_sync.set()
 
-                    frame = AxiStreamFrame([], [], [], [], [])
-
-            await RisingEdge(self.clock)
-
-            if hasattr(self.bus, "tready"):
-                self.bus.tready <= (not self.full() and not self.pause)
-
-    async def _run_pause(self):
-        for val in self._pause_generator:
-            self.pause = val
-            await RisingEdge(self.clock)
+                    frame = None
 
 
-class AxiStreamMonitor(object):
+class AxiStreamSink(AxiStreamMonitor, AxiStreamPause):
 
-    _signals = ["tdata"]
-    _optional_signals = ["tvalid", "tready", "tlast", "tkeep", "tid", "tdest", "tuser"]
+    _type = "sink"
 
-    def __init__(self, entity, name, clock, reset=None, byte_size=None, byte_lanes=None, *args, **kwargs):
-        self.log = logging.getLogger(f"cocotb.{entity._name}.{name}")
-        self.entity = entity
-        self.clock = clock
-        self.reset = reset
-        self.bus = Bus(self.entity, name, self._signals, optional_signals=self._optional_signals, **kwargs)
+    _init_x = False
 
-        self.log.info("AXI stream monitor")
-        self.log.info("cocotbext-axi version %s", __version__)
-        self.log.info("Copyright (c) 2020 Alex Forencich")
-        self.log.info("https://github.com/alexforencich/cocotbext-axi")
+    _valid_init = None
+    _ready_init = 0
 
-        super().__init__(*args, **kwargs)
+    def __init__(self, bus, clock, reset=None, reset_active_level=True,
+            byte_size=None, byte_lanes=None, *args, **kwargs):
 
-        self.active = False
-        self.queue = deque()
-        self.sync = Event()
-        self.read_queue = []
+        super().__init__(bus, clock, reset, reset_active_level, byte_size, byte_lanes, *args, **kwargs)
 
-        self.queue_occupancy_bytes = 0
-        self.queue_occupancy_frames = 0
+        self.queue_occupancy_limit_bytes = -1
+        self.queue_occupancy_limit_frames = -1
 
-        self.width = len(self.bus.tdata)
-        self.byte_lanes = 1
-
-        self.reset = reset
-
-        if hasattr(self.bus, "tkeep"):
-            self.byte_lanes = len(self.bus.tkeep)
-            if byte_size is not None or byte_lanes is not None:
-                raise ValueError("Cannot specify byte_size or byte_lanes if tkeep is connected")
+    def full(self):
+        if self.queue_occupancy_limit_bytes > 0 and self.queue_occupancy_bytes > self.queue_occupancy_limit_bytes:
+            return True
+        elif self.queue_occupancy_limit_frames > 0 and self.queue_occupancy_frames > self.queue_occupancy_limit_frames:
+            return True
         else:
-            if byte_lanes is not None:
-                self.byte_lanes = byte_lanes
-                if byte_size is not None:
-                    raise ValueError("Cannot specify both byte_size and byte_lanes")
-            elif byte_size is not None:
-                self.byte_lanes = self.width // byte_size
+            return False
 
-        self.byte_size = self.width // self.byte_lanes
-        self.byte_mask = 2**self.byte_size-1
+    def _handle_reset(self, state):
+        super()._handle_reset(state)
 
-        self.log.info("AXI stream monitor configuration:")
-        self.log.info("  Byte size: %d bits", self.byte_size)
-        self.log.info("  Data width: %d bits (%d bytes)", self.width, self.byte_lanes)
-        self.log.info("  tvalid: %s", "present" if hasattr(self.bus, "tvalid") else "not present")
-        self.log.info("  tready: %s", "present" if hasattr(self.bus, "tready") else "not present")
-        self.log.info("  tlast: %s", "present" if hasattr(self.bus, "tlast") else "not present")
-        if hasattr(self.bus, "tkeep"):
-            self.log.info("  tkeep width: %d bits", len(self.bus.tkeep))
-        else:
-            self.log.info("  tkeep: not present")
-        if hasattr(self.bus, "tid"):
-            self.log.info("  tid width: %d bits", len(self.bus.tid))
-        else:
-            self.log.info("  tid: not present")
-        if hasattr(self.bus, "tdest"):
-            self.log.info("  tdest width: %d bits", len(self.bus.tdest))
-        else:
-            self.log.info("  tdest: not present")
-        if hasattr(self.bus, "tuser"):
-            self.log.info("  tuser width: %d bits", len(self.bus.tuser))
-        else:
-            self.log.info("  tuser: not present")
-
-        if self.byte_lanes * self.byte_size != self.width:
-            raise ValueError(f"Bus does not evenly divide into byte lanes "
-                f"({self.byte_lanes} * {self.byte_size} != {self.width})")
-
-        cocotb.fork(self._run())
-
-    async def recv(self, compact=True):
-        while self.empty():
-            self.sync.clear()
-            await self.sync.wait()
-        return self.recv_nowait(compact)
-
-    def recv_nowait(self, compact=True):
-        if self.queue:
-            frame = self.queue.popleft()
-            self.queue_occupancy_bytes -= len(frame)
-            self.queue_occupancy_frames -= 1
-            if compact:
-                frame.compact()
-            return frame
-        return None
-
-    async def read(self, count=-1):
-        while not self.read_queue:
-            frame = await self.recv(compact=True)
-            self.read_queue.extend(frame.tdata)
-        return self.read_nowait(count)
-
-    def read_nowait(self, count=-1):
-        while not self.empty():
-            frame = self.recv_nowait(compact=True)
-            self.read_queue.extend(frame.tdata)
-        if count < 0:
-            count = len(self.read_queue)
-        data = self.read_queue[:count]
-        del self.read_queue[:count]
-        return data
-
-    def count(self):
-        return len(self.queue)
-
-    def empty(self):
-        return not self.queue
-
-    def idle(self):
-        return not self.active
-
-    async def wait(self, timeout=0, timeout_unit='ns'):
-        if not self.empty():
-            return
-        self.sync.clear()
-        if timeout:
-            await First(self.sync.wait(), Timer(timeout, timeout_unit))
-        else:
-            await self.sync.wait()
+        if hasattr(self.bus, "tready"):
+            self.bus.tready <= 0
 
     async def _run(self):
-        frame = AxiStreamFrame([], [], [], [], [])
+        frame = None
         self.active = False
 
         while True:
-            await ReadOnly()
+            await RisingEdge(self.clock)
 
             # read handshake signals
             tready_sample = (not hasattr(self.bus, "tready")) or self.bus.tready.value
             tvalid_sample = (not hasattr(self.bus, "tvalid")) or self.bus.tvalid.value
 
-            if self.reset is not None and self.reset.value:
-                await RisingEdge(self.clock)
-                frame = AxiStreamFrame([], [], [], [], [])
-                self.active = False
-                continue
-
             if tready_sample and tvalid_sample:
+                if frame is None:
+                    if self.byte_size == 8:
+                        frame = AxiStreamFrame(bytearray(), [], [], [], [])
+                    else:
+                        frame = AxiStreamFrame([], [], [], [], [])
+                    frame.sim_time_start = get_sim_time()
+
                 for offset in range(self.byte_lanes):
 
                     frame.tdata.append((self.bus.tdata.value.integer >> (offset * self.byte_size)) & self.byte_mask)
@@ -802,14 +678,16 @@ class AxiStreamMonitor(object):
                         frame.tuser.append(self.bus.tuser.value.integer)
 
                 if not hasattr(self.bus, "tlast") or self.bus.tlast.value:
-                    if self.byte_size == 8:
-                        frame.tdata = bytearray(frame.tdata)
-
+                    frame.sim_time_end = get_sim_time()
                     self.log.info("RX frame: %s", frame)
 
+                    self.queue_occupancy_bytes += len(frame)
+                    self.queue_occupancy_frames += 1
+
                     self.queue.append(frame)
-                    self.sync.set()
+                    self.queue_sync.set()
 
-                    frame = AxiStreamFrame([], [], [], [], [])
+                    frame = None
 
-            await RisingEdge(self.clock)
+            if hasattr(self.bus, "tready"):
+                self.bus.tready <= (not self.full() and not self.pause)
